"""Vaillant operation modes."""
from typing import Optional
import attr


@attr.s
class Mode:
    """This is the base class for modes, it groups :class:`QuickVeto`,
    :class:`OperatingMode`, :class:`QuickMode` and :class:`SettingMode`.

    Args:
        name (str): The name of the mode.
    """

    name = attr.ib(type=str, default=None)


@attr.s(frozen=True)
class OperatingMode(Mode):
    """Represents the operating mode of a
    :class:`~pymultimatic.model.component.Component`"""


class OperatingModes:
    """Groups all :class:`OperatingMode` together."""

    AUTO = OperatingMode('AUTO')
    """The automatic mode controls the heating circuit in accordance with the
    set desired temperature and the set time slots.
    """

    DAY = OperatingMode('DAY')
    """The Day operating mode brings the heating circuit to the desired Day
    temperature set, without taking time periods into account.
    """

    NIGHT = OperatingMode('NIGHT')
    """The Set-back (Night) operating mode brings the heating circuit to the
    desired Set-back temperature set, without taking time periods into
    consideration.
    """

    ON = OperatingMode('ON')
    """The operating mode "On" applies to
    :class:`~pymultimatic.model.component.Hotwater` and
    :class:`~pymultimatic.model.component.Circulation`, it means the
    :class:`~pymultimatic.model.component.Component` is running on.
    """

    OFF = OperatingMode('OFF')
    """The operating mode "Off" switches the corresponding function off. The
    frost protection function is activated.
    """

    MANUAL = OperatingMode('MANUAL')
    """The manual operating mode is used by a
    :class:`~pymultimatic.model.component.Room` to set manually the target
    temperature and keep it until next manual change.
    """

    QUICK_VETO = OperatingMode('QUICK_VETO')
    """The Quick veto function can be used to make one-off changes to the
    temperature. This function sets the desired target temperature, which may
    differ from that configured in the
    :class:`~pymultimatic.model.timeprogram.TimeProgram` currently being used
    . The newly set temperature will be applied for a certain period of
    time."""

    _VALUES = {opm.name: opm for opm in [AUTO, DAY, NIGHT, ON, OFF, MANUAL,
                                         QUICK_VETO]}

    @classmethod
    def get(cls, name: str) -> OperatingMode:
        """Get :class:`OperatingMode` by name.

        Args:
            name (str): Name of the operating mode

        Returns:
            The corresponding operating mode or None
        """
        return cls._VALUES[name]


@attr.s
class QuickVeto(Mode):
    """Represents a quick veto which can be applied to a
    :class:`~pymultimatic.model.component.Zone` or a
    :class:`~pymultimatic.model.component.Room`.

    For a :class:`~pymultimatic.model.component.Room`, quick veto duration is
    customizable and it's possible to get the remaining duration through the
    API. The max value is 1440 minutes (1 day)

    For a :class:`~pymultimatic.model.component.Zone`, quick veto duration is
    **not** customizable (default duration is 6 hours) and the API returns
    always *0*.

    Args:
        duration (int): remaining duration in minutes
            before the quick veto expires. Since remaining duration is not
            available for a :class:`~pymultimatic.model.component.Zone`,
            value will be `None`.
        target (float): target temperature defined by the quick
            veto.
    """

    name = attr.ib(default=OperatingModes.QUICK_VETO.name, init=False)
    duration = attr.ib(type=Optional[int], default=None)
    target = attr.ib(type=float, default=None)

    # pylint: disable=no-self-use
    @duration.validator
    def _duration_validator(self, attribute: str,
                            value: Optional[int]) -> None:
        if value is not None and value > 1440:
            raise ValueError('{} with value {} is not valid'
                             .format(attribute, value))

    # pylint: disable=no-self-use
    @target.validator
    def _target_validator(self, attribute: str,
                          value: Optional[float]) -> None:
        if value is None or not 5 <= value <= 30:
            raise ValueError('{} with value {} is not valid'
                             .format(attribute, value))


@attr.s(frozen=True)
class SettingMode(Mode):
    """This is the setting which is configured in
    :class:`~pymultimatic.model.timeprogram.TimePeriodSetting`."""


class SettingModes:
    """Groups all :class:`SettingMode` together."""

    ON = SettingMode('ON')
    """Within the time period,
    :class:`~pymultimatic.model.component.HotWater` generation is switched on
    and maintains the hot water in the domestic hot water cylinder at the
    preset temperature. This also means the
    :class:`~pymultimatic.model.component.Circulation` is switched off.
    """

    OFF = SettingMode('OFF')
    """Outside the time period,
    :class:`~pymultimatic.model.component.HotWater` generation (and
    :class:`~pymultimatic.model.component.Circulation`) is switched off.
    """

    DAY = SettingMode('DAY')
    """Within the time periods, the controller brings the
    :class:`~pymultimatic.model.component.Zone` temperature to the `DAY`
    temp. heating that has been set (Comfort mode).
    """

    NIGHT = SettingMode('NIGHT')
    """Outside of the time period, the controller brings the
    :class:`~pymultimatic.model.component.Zone` temperature to the Set-back
    temp. heating (`NIGHT`) that has been set.
    """

    _VALUES = {sm.name: sm for sm in [ON, OFF, DAY, NIGHT]}

    @classmethod
    def get(cls, name: str) -> SettingMode:
        """Get :class:`SettingMode` by name.

        Args:
            name (str): Name of the setting mode

        Returns:
            The corresponding setting mode or None
        """
        return cls._VALUES[name]


@attr.s
class ActiveMode:
    """Active mode will let you know the real target temperature and the real
    :class:`Mode` applied to a
    :class:`~pymultimatic.model.component.Component`.

    Note:
        This is based on the interpretation of the data coming from the API.

    It can be quite complex to know what is going on from a
    :class:`~pymultimatic.model.component.Component`,
    the API returns data at different places like for example:

    * the :class:`OperatingMode` (and the :class:`QuickVeto`) for a
      :class:`~pymultimatic.model.component.Component` are found at one place

    * the system wise :class:`QuickMode` is found somewhere else

    * not all the :class:`QuickMode` are applicable to all
      :class:`~pymultimatic.model.component.Component`

    So, if a :class:`~pymultimatic.model.component.Zone` is configured with
    :class:`OperatingModes.AUTO` mode and there is a
    :class:`QuickModes.QM_SYSTEM_OFF` active, the API won't tell you what is
    the real mode applied to that :class:`~pymultimatic.model.component.Zone`
    . In this case, the :class:`QuickModes.QM_SYSTEM_OFF` will actually take
    precedence and it will have impact on the target temperature.

    Args:
        target (Optional[float]): Target temperature the
            :class:`~pymultimatic.model.component.Component` tries to reach.
            Please note, there will be no `target_temperature` for
            :class:`~pymultimatic.model.component.Circulation`.
        current (Mode): This is the mode that will be applied by your
            installation.
        sub: (Mode): If `mode` is :class:`OperatingModes.AUTO`,
            then `sub_mode` will be populated with the actual configured mode,
            coming from
            :class:`~pymultimatic.model.timeprogram.TimeProgramDaySetting.mode`
            , otherwise it will be `None`
    """

    target = attr.ib(type=Optional[float])
    current = attr.ib(type=Mode)
    sub = attr.ib(type=Optional[Mode], default=None)



"""Groups everything related to a zone."""
from typing import Optional

import attr

from . import (Function, ActiveMode, OperatingModes, constants, Component)


class ZoneHeating(Function):
    """Represent the heating function of a zone."""

    def _active_mode(self) -> ActiveMode:
        if self.operating_mode == OperatingModes.OFF:
            mode = ActiveMode(constants.FROST_PROTECTION_TEMP,
                              OperatingModes.OFF)
        elif self.operating_mode == OperatingModes.DAY:
            mode = ActiveMode(self.target_high,
                              OperatingModes.DAY)
        else:  # MODE_NIGHT
            mode = ActiveMode(self.target_low,
                              OperatingModes.NIGHT)
        return mode


class ZoneCooling(Function):
    """Represent the cooling function of a zone."""

    def _active_mode(self) -> ActiveMode:
        if self.operating_mode == OperatingModes.OFF:
            mode = ActiveMode(self.target_low, OperatingModes.OFF)
        else:  # MODE ON
            mode = ActiveMode(self.target_high, OperatingModes.ON)
        return mode


@attr.s
class Zone(Component):
    """This is representing a zone from the system.

     Note:
         If `rbr` (Room By Room) is True, the zone itself doesn't mean anything
         anymore, it means rooms are 'controlling' the zone.

     Args:
         active_function (str): Indicate what the zone is doing (basically
             'HEATING' or 'STANDBY').
         rbr (bool): Room By Room, means the zone is controlled by ambisense.
         heating (ZoneHeating): Heating function of the zone.
         cooling (ZoneCooling): Cooling function of the zone.
     """

    MODES = [OperatingModes.AUTO, OperatingModes.OFF, OperatingModes.DAY,
             OperatingModes.NIGHT, OperatingModes.QUICK_VETO]
    """List of mode that are applicable to zones component."""

    MIN_TARGET_TEMP = constants.FROST_PROTECTION_TEMP
    """Min temperature that can be apply to a zone."""

    MAX_TARGET_TEMP = constants.THERMOSTAT_MAX_TEMP
    """Max temperature that can be apply to a zone."""

    active_function = attr.ib(type=str, default=None)
    rbr = attr.ib(type=bool, default=False)
    heating = attr.ib(type=ZoneHeating, default=None)
    cooling = attr.ib(type=Optional[ZoneCooling], default=None)

    @property
    def active_mode(self) -> ActiveMode:
        if self.quick_veto:
            return ActiveMode(self.quick_veto.target,
                              OperatingModes.QUICK_VETO)
        return self.heating.active_mode



"""Groups everything related to live report sensor."""
import attr


@attr.s
class Report:
    """Represent a live report sensor."""

    # pylint: disable=invalid-name
    id = attr.ib(type=str)
    value = attr.ib(type=float)
    name = attr.ib(type=str)
    unit = attr.ib(type=str)
    device_id = attr.ib(type=str)
    device_name = attr.ib(type=str)



"""Low level connector module."""
import logging
from typing import Any, Dict, Optional

import attr
import aiohttp
from yarl import URL

from . import ApiError, urls, defaults

_LOGGER = logging.getLogger('Connector')

HEADER = {'content-type': 'application/json'}


@attr.s
class Connector:
    """This is the low level smart.vaillant.com API connector.

    This is returning the raw JSON from responses or an
    :exc:`~pymultimatic.api.error.ApiError` if something goes wrong
    (basically, when response error code is 4xx or 5xx).

    On the first call, the connector will login automatically (if
    :func:`~login` was not called previously).

    On following calls, the connector will re-use the cookie received from the
    API. If the connector receives an HTTP 401, it will clear the cookie and
    try to re-login before raising an `~pymultimatic.api.error.ApiError`. This
    also means the connector is able to reconnect automatically when cookie is
    outdated.

    Please use :mod:`~pymultimatic.api.urls` in order to generate URL to be
    passed to the connector.64:36

    Args:
        user (str): User to login with.
        password (str): Password associated with the user.
        session: (aiohttp.ClientSession): Session.
        smartphone_id (str): This is required by the API to login.
    """

    _user = attr.ib(type=str)
    _password = attr.ib(type=str, repr=False)
    _session = attr.ib(type=aiohttp.ClientSession)
    _smartphone_id = attr.ib(type=str, default=defaults.SMARTPHONE_ID)

    async def login(self, force: bool = False) -> bool:
        """Log in to the API.

        By default, the ``connector`` will try to re-use cookies

        Args:
            force (bool): If set to ``True``, the connector will clear
                the cookie (if any) and start a new authentication, otherwise
                it will re-use the existing cookie.

        Returns:
            bool: True/False if authentication succeeded or not.
        """
        if force:
            self._clear_cookies()

        if self._get_cookies():
            return True

        token = await self._token()
        await self._authenticate(token)
        return True

    async def is_logged(self) -> bool:
        """Check if the connector is already logged in.

        It relies in the presence of cookies.
        """
        return len(self._get_cookies()) > 0

    async def logout(self) -> bool:
        """Get logged out of the API.

        It first sends a ``logout`` request to the API (it means cookies are
        invalidated), then cookies will be cleared, regardless of the result
        of the ``logout`` request.

        The connector will have to request a new token and ask for cookies if
        a new request is done.


        Raises:
            ApiError: When something went wrong with the API call.
        """
        try:
            if self._get_cookies():
                await self.post(urls.logout())
            return True
        finally:
            self._clear_cookies()

    async def _token(self) -> str:
        params = {
            "smartphoneId": self._smartphone_id,
            "username": self._user,
            "password": self._password
        }

        token_res = await self._session.post(url=urls.new_token(),
                                             json=params,
                                             headers=HEADER)
        if token_res.status == 200:
            json = await token_res.json()
            return str(json['body']['authToken'])
        raise ApiError('Login/password invalid', response=token_res)

    async def _authenticate(self, token: str) -> None:
        params = {
            "smartphoneId": self._smartphone_id,
            "username": self._user,
            "authToken": token
        }

        auth_res = await self._session.post(url=urls.authenticate(),
                                            json=params,
                                            headers=HEADER)

        if auth_res.status > 399:
            raise ApiError("Unable to authenticate", response=auth_res)

    def _get_cookies(self) -> Dict[Any, Any]:
        return self._session.cookie_jar.filter_cookies(URL(urls.base()))

    def _clear_cookies(self) -> None:
        self._session.cookie_jar.clear()

    async def get(self, url: str,
                  payload: Optional[Dict[str, Any]] = None) -> Any:
        """Do a get against vaillant API."""
        return await self.request('get', url, payload)

    async def delete(self, url: str,
                     payload: Optional[Dict[str, Any]] = None) -> Any:
        """Do a delete against vaillant API."""
        return await self.request('delete', url, payload)

    async def put(self, url: str,
                  payload: Optional[Dict[str, Any]] = None) -> Any:
        """Do a put against vaillant API."""
        return await self.request('put', url, payload)

    async def post(self, url: str,
                   payload: Optional[Dict[str, Any]] = None) -> Any:
        """Do a post against vaillant API."""
        return await self.request('post', url, payload)

    async def request(self, method: str, url: str,
                      payload: Optional[Dict[str, Any]] = None) -> Any:
        """Do a request against vaillant API."""
        async with self._session.request(method, url, data=payload) as resp:
            if resp.status == 401:
                await self.login(True)
                return await self.request(method, url, payload)

            if resp.status > 399:
                raise ApiError('Cannot ' + method + ' ' + url, response=resp)

            return await resp.json()



"""Mapped model from the API."""
# pylint: disable=cyclic-import
from .mode import Mode, OperatingMode, OperatingModes, QuickVeto, ActiveMode,\
    SettingMode, SettingModes
from .timeprogram import TimeProgram, TimeProgramDay, TimePeriodSetting
from .common import Component, Function
from .zone import ZoneCooling, ZoneHeating, Zone
from .room import Device, Room
from .status import BoilerStatus, Error
from .syncstate import SyncState
from .dhw import Dhw, HotWater, Circulation
from .report import Report
from .quick_mode import QuickMode, QuickModes, HolidayMode
from .ventilation import Ventilation
from .system import System, SystemInfo



"""Convenient manager to easily gets data from API."""
import logging
from datetime import date, timedelta
from typing import Optional, List, Callable, Any

from aiohttp import ClientSession

from .api import Connector, urls, payloads, defaults, ApiError
from .model import mapper, System, HotWater, QuickMode, QuickVeto, Room, \
    Zone, OperatingMode, Circulation, OperatingModes, constants

_LOGGER = logging.getLogger('SystemManager')


# pylint: disable=too-many-public-methods
class SystemManager:
    """This is a convenient manager to help interact with vaillant API.

    The manager is calling the
    :class:`~pymultimatic.api.connector.ApiConnector`, the manager will
    provide to correct :mod:`~pymultimatic.api.payloads` and
    :mod:`~pymultimatic.api.urls` according to what you want to do.

    The manager is throwing :class:`~pymultimatic.api.error.ApiError`
    without altering it.

    Args:
        user (str): User to login with.
        password (str): Password associated with the user.
        smartphone_id (str): This is required by the API to login.
    """
    # pylint: disable=too-many-arguments
    def __init__(self,
                 user: str,
                 password: str,
                 session: ClientSession,
                 smartphone_id: str = defaults.SMARTPHONE_ID,
                 serial: Optional[str] = None):
        self._connector: Connector = Connector(
            user,
            password,
            session,
            smartphone_id)
        self._serial = serial
        self._fixed_serial = self._serial is not None

    async def login(self, force_login: bool = False) -> bool:
        """Try to login to the API, see
        :func:`~pymultimatic.api.connector.ApiConnector.login`

        Args:
            force_login (bool): Whether the login should be forced or not.

        Returns:
            bool: True/False if authentication succeeded or not.
        """
        return await self._connector.login(force_login)

    # pylint: disable=too-many-locals
    async def get_system(self) -> System:
        """Get the full :class:`~pymultimatic.model.system.System`. It may
        take some times, it actually does multiples API calls, depending on
        your system configuration.

        Returns:
            System: the full system.
        """

        facilities_req = self._call_api(urls.facilities_list)
        full_system_req = self._call_api(urls.system)
        live_report_req = self._call_api(urls.live_report)
        hvac_state_req = self._call_api(urls.hvac)
        gateway_req = self._call_api(urls.gateway_type)

        gateway = await gateway_req
        facilities = await facilities_req
        hvac_state = await hvac_state_req
        system_info = mapper.map_system_info(facilities, gateway, hvac_state)

        boiler_status = mapper.map_boiler_status(hvac_state)
        errors = mapper.map_errors(hvac_state)

        full_system = await full_system_req
        holiday = mapper.map_holiday_mode(full_system)
        zones = mapper.map_zones(full_system)
        outdoor_temp = mapper.map_outdoor_temp(full_system)
        quick_mode = mapper.map_quick_mode(full_system)
        ventilation = mapper.map_ventilation(full_system)

        live_report = await live_report_req
        dhw = mapper.map_dhw(full_system, live_report)
        reports = mapper.map_reports(live_report)

        rooms: List[Room] = []
        if [z for z in zones if z.rbr]:
            rooms_req = self._call_api(urls.rooms)
            rooms_raw = await rooms_req
            rooms = mapper.map_rooms(rooms_raw)

        return System(holiday=holiday,
                      quick_mode=quick_mode,
                      info=system_info,
                      zones=zones,
                      rooms=rooms,
                      dhw=dhw,
                      reports=reports,
                      outdoor_temperature=outdoor_temp,
                      boiler_status=boiler_status,
                      errors=errors,
                      ventilation=ventilation)

    async def get_hot_water(self, dhw_id: str) -> Optional[HotWater]:
        """Get the :class:`~pymultimatic.model.component.HotWater`
        information for the given id.

        Args:
            dhw_id (str): domestic hot water id (same as
                :func:`get_circulation`).

        Returns:
            HotWater: the hot water information, if any.
        """
        dhw_req = self._call_api(urls.hot_water,
                                 params={'id': dhw_id})
        lv_req = self._call_api(urls.live_report)

        report = await lv_req
        dhw = await dhw_req
        return mapper.map_hot_water_alone(dhw, dhw_id, report)

    async def get_room(self, room_id: str) -> Optional[Room]:
        """Get the :class:`~pymultimatic.model.component.Room` information
        for the given id.

        Args:
            room_id (str): Id of the room, this is actually an index, 1,2,3,4
            ... depending of the number of rooms you have.

        Returns:
            Room: the room information, if any.
        """
        new_room = await self._call_api(urls.room, params={'id': room_id})
        return mapper.map_room(new_room)

    async def get_zone(self, zone_id: str) -> Optional[Zone]:
        """"Get the :class:`~pymultimatic.model.component.Zone` information
        for the given id.

        Args:
            zone_id (str): Name of the room, basically, you set it through the
                VRC 700.

        Returns:
            Zone: the zone information, if any.
        """
        new_zone = await self._call_api(urls.zone, params={'id': zone_id})
        return mapper.map_zone(new_zone)

    async def get_circulation(self, dhw_id: str) -> Optional[Circulation]:
        """"Get the :class:`~pymultimatic.model.component.Circulation`
        information for the given id.

        Args:
            dhw_id (str): domestic hot water id (same as :func:`get_hot_water`)

        Returns:
            Circulation: the circulation information, if any.
        """
        new_circulation = await self._call_api(urls.circulation,
                                               params={'id': dhw_id})
        return mapper.map_circulation_alone(new_circulation, dhw_id)

    async def set_hot_water_setpoint_temperature(self, dhw_id: str,
                                                 temperature: float) -> None:
        """This set the target temperature for *hot water*."""
        _LOGGER.debug("Will set dhw target temperature to %s", temperature)

        payload = payloads \
            .hotwater_temperature_setpoint(self._round(temperature))

        await self._call_api(
            urls.hot_water_temperature_setpoint,
            params={'id': dhw_id},
            payload=payload
        )

    async def set_hot_water_operating_mode(self, dhw_id: str,
                                           new_mode: OperatingMode) -> None:
        """Set new operating mode for
        :class:`~pymultimatic.model.component.HotWater`. The mode should be
        listed here :class:`~pymultimatic.model.component.HotWater.MODES`
        otherwise it won't have any effect.


        Note:
            To set :class:`~pymultimatic.model.mode.QuickMode`, you
            have to use :func:`set_quick_mode`.

        Note:
            This call won't have any effect if there is a
            :class:`~pymultimatic.model.mode.QuickMode` activated, if you
            want to remove the quick mode, use :func:`remove_quick_mode`.

        Args:
            dhw_id (str): domestic hot water id.
            new_mode (OperatingMode): The new mode to set.
        """
        _LOGGER.debug("Will try to set hot water mode to %s", new_mode)

        if new_mode in HotWater.MODES:
            _LOGGER.debug("New mode is %s", new_mode)
            await self._call_api(
                urls.hot_water_operating_mode,
                params={'id': dhw_id},
                payload=payloads.hot_water_operating_mode(new_mode.name)
            )
        else:
            _LOGGER.debug("New mode is not available for hot water %s",
                          new_mode)

    async def set_room_operating_mode(self, room_id: str,
                                      new_mode: OperatingMode) -> None:
        """Set new operating mode for
        :class:`~pymultimatic.model.component.Room`. The mode should be
        listed here :class:`~pymultimatic.model.component.Room.MODES`
        otherwise it won't have any effect.


        Note:
            To set :class:`~pymultimatic.model.mode.QuickMode`, you
            have to use :func:`set_quick_mode`.

        Note:
            This call won't have any effect if there is a
            :class:`~pymultimatic.model.mode.QuickMode` activated, if you
            want to remove the quick mode, use :func:`remove_quick_mode`.

        Note:
            To set :class:`~pymultimatic.model.mode.QuickVeto`, you have to
            use :func:`set_room_quick_veto`

        Args:
            room_id (str): id of the room.
            new_mode (OperatingMode): The new mode to set.
        """
        if new_mode in Room.MODES and new_mode != OperatingModes.QUICK_VETO:
            _LOGGER.debug("New mode is %s", new_mode)
            await self._call_api(
                urls.room_operating_mode,
                params={'id': room_id},
                payload=payloads.room_operating_mode(new_mode.name)
            )
        else:
            _LOGGER.debug("mode is not available for room %s", new_mode)

    async def set_zone_operating_mode(self, zone_id: str,
                                      new_mode: OperatingMode) -> None:
        """Set new operating mode for
        :class:`~pymultimatic.model.component.Zone`. The mode should be
        listed here :class:`~pymultimatic.model.component.Zone.MODES`
        otherwise it won't have any effect.


        Note:
            To set :class:`~pymultimatic.model.mode.QuickMode`, you
            have to use :func:`set_quick_mode`.

        Note:
            This call won't have any effect if there is a
            :class:`~pymultimatic.model.mode.QuickMode` activated, if you
            want to remove the quick mode, use :func:`remove_quick_mode`.

        Note:
            To set :class:`~pymultimatic.model.mode.QuickVeto`, you have to
            use :func:`set_zone_quick_veto`

        Args:
            zone_id (str): id of the zone.
            new_mode (OperatingMode): The new mode to set.
        """
        if new_mode in Zone.MODES and new_mode != OperatingModes.QUICK_VETO:
            _LOGGER.debug("New mode is %s", new_mode)
            await self._call_api(
                urls.zone_heating_mode,
                params={'id': zone_id},
                payload=payloads.zone_operating_mode(new_mode.name)
            )
        else:
            _LOGGER.debug("mode is not available for zone %s", new_mode)

    async def set_quick_mode(self, quick_mode: QuickMode) -> None:
        """Set a :class:`~pymultimatic.model.mode.QuickMode` system wise.

        This will override the current
        :class:`~pymultimatic.model.mode.QuickMode`, if any.

        Args:
            quick_mode (QuickMode): the quick mode to set, see
                :class:`~pymultimatic.model.mode.QuickModes`
        """
        await self._call_api(
            urls.system_quickmode,
            payload=payloads.quickmode(quick_mode.name)
        )

    async def remove_quick_mode(self) -> None:
        """Removes current :class:`~pymultimatic.model.mode.QuickMode`.

        Note:
            if there is not :class:`~pymultimatic.model.mode.QuickMode` set,
            the API returns an error (HTTP 409). **This error is swallowed by
            the manager**, so you don't have to handle it."""
        try:
            await self._call_api(urls.system_quickmode, 'delete')
        except ApiError as exc:
            if exc.response is None or exc.response.status != 409:
                raise exc

    async def set_room_quick_veto(self, room_id: str,
                                  quick_veto: QuickVeto) -> None:
        """Set a :class:`~pymultimatic.model.mode.QuickVeto` for a
        :class:`~pymultimatic.model.component.Room`.
        It will override the current
        :class:`~pymultimatic.model.mode.QuickVeto`, if any.

        Args:
            room_id (str): Id of the room.
            quick_veto (QuickVeto): Quick veto to set.
        """
        payload = payloads.room_quick_veto(
            self._round(quick_veto.target),
            quick_veto.duration
        )
        await self._call_api(
            urls.room_quick_veto,
            params={'id': room_id},
            payload=payload
        )

    async def remove_room_quick_veto(self, room_id: str) -> None:
        """Remove the :class:`~pymultimatic.model.mode.QuickVeto` from a
        :class:`~pymultimatic.model.component.Room`.

        Args:
            room_id (str): Id of the room.
        """

        await self._call_api(
            urls.room_quick_veto,
            'delete',
            params={'id': room_id}
        )

    async def set_zone_quick_veto(self, zone_id: str, quick_veto: QuickVeto) \
            -> None:
        """Set a :class:`~pymultimatic.model.mode.QuickVeto` for a
        :class:`~pymultimatic.model.component.Zone`.
        It will override the current
        :class:`~pymultimatic.model.mode.QuickVeto`, if any.

        Args:
            zone_id (str): Id of the zone.
            quick_veto (QuickVeto): Quick veto to set.
        """
        payload = payloads.zone_quick_veto(
            self._round(quick_veto.target))

        await self._call_api(
            urls.zone_quick_veto,
            params={'id': zone_id},
            payload=payload
        )

    async def remove_zone_quick_veto(self, zone_id: str) -> None:
        """Remove the :class:`~pymultimatic.model.mode.QuickVeto` from a
        :class:`~pymultimatic.model.component.Zone`.

        Args:
            zone_id (str): Id of the zone.
        """
        await self._call_api(
            urls.zone_quick_veto,
            'delete',
            params={'id': zone_id}
        )

    async def set_room_setpoint_temperature(self, room_id: str,
                                            temperature: float) -> None:
        """Set the new current target temperature for a
        :class:`~pymultimatic.model.component.Room`.

        Note:
            Setting the target temperature this way will actually set the room
            operating mode to
            :class:`~pymultimatic.model.mode.OperatingModes.MANUAL` (and
            target temperature to the desired temperature). It means the target
            temperature will remain the same until you change the target
            temperature again or the operating mode.

        Args:
            room_id (str): Id of the room.
            temperature (float): Target temperature to set.
        """

        _LOGGER.debug("Will try to set room target temperature to %s",
                      temperature)

        await self._call_api(
            urls.room_temperature_setpoint,
            params={'id': room_id},
            payload=payloads.room_temperature_setpoint(
                self._round(temperature))
        )

    async def set_zone_setpoint_temperature(self, zone_id: str,
                                            temperature: float) -> None:
        """Set the configured temperature for the
        :class:`~pymultimatic.model.mode.SettingModes.DAY` mode.

        Note:
            It won't alter the
            :class:`~pymultimatic.model.mode.OperatingMode`.

        Args:
            zone_id (str): Id of the zone.
            temperature (float): New temperature.
        """
        _LOGGER.debug("Will try to set zone target temperature to %s",
                      temperature)

        payload = payloads.zone_temperature_setpoint(self._round(temperature))

        await self._call_api(
            urls.zone_heating_setpoint_temperature,
            params={'id': zone_id},
            payload=payload
        )

    async def set_zone_setback_temperature(self, zone_id: str,
                                           temperature: float) -> None:
        """Set the configured temperature for the
        :class:`~pymultimatic.model.mode.SettingModes.NIGHT` mode.

        Note:
            It won't alter the
            :class:`~pymultimatic.model.mode.OperatingMode`.

        Args:
            zone_id (str): Id of the zone.
            temperature (float): New temperature.
        """
        _LOGGER.debug("Will try to set zone setback temperature to %s",
                      temperature)

        await self._call_api(
            urls.zone_heating_setback_temperature,
            params={'id': zone_id},
            payload=payloads.zone_temperature_setback(self._round(temperature))
        )

    async def set_holiday_mode(self, start_date: date, end_date: date,
                               temperature: float) -> None:
        """Set the :class:`~pymultimatic.model.mode.HolidayMode`.

        Args:
            start_date (date): Start date of the holiday mode.
            end_date (date): End date of the holiday mode.
            temperature (float): Target temperature while holiday mode
                :class:`~pymultimatic.model.mode.HolidayMode.is_applied`
        """
        payload = payloads.holiday_mode(
            True,
            start_date,
            end_date,
            self._round(temperature)
        )

        await self._call_api(
            urls.system_holiday_mode,
            payload=payload
        )

    async def remove_holiday_mode(self) -> None:
        """Remove :class:`~pymultimatic.model.mode.HolidayMode`.

        Note:
            There is a little workaround here, since the API doesn't simply
            accept a DELETE request to remove the
            :class:`~pymultimatic.model.mode.HolidayMode`, so the manager is
            setting:

            * the start date to two days before

            * the end date to yesterday

            * temperature to frost protection

            * active to `False`

            This will ensure the
            :class:`~pymultimatic.model.mode.HolidayMode` is not active.

        """

        payload = payloads.holiday_mode(
            False,
            date.today() - timedelta(days=2),
            date.today() - timedelta(days=1),
            constants.FROST_PROTECTION_TEMP
        )

        await self._call_api(
            urls.system_holiday_mode,
            payload=payload
        )

    async def request_hvac_update(self) -> None:
        """Request an hvac update. This allow the vaillant API to read the data
        from your system.

        This is necessary to update
        :class:`~pymultimatic.model.status.BoilerStatus` and
        :class:`~pymultimatic.model.status.Error`

        Note:
            The **request** done by the manager is done **synchronously**,
            but the **update** processing (basically, reading data from your
            system) is done **asynchronously** by vaillant API.

        Note:
            The manager will try to reduce number of call, I think the call is
            costly. If the
            :class:`~pymultimatic.model.syncstate.SyncState`
            :class:`~pymultimatic.model.syncstate.SyncState.is_pending`,
            the manager will skip the call to the API.


        it can take some times for the update to occur (Most of the
        time, it takes about 1 or 2 minutes before you can see changes, if any)

        It the request is done too often, the API may return an error
        (HTTP 409).

        """

        state = mapper.map_hvac_sync_state(await self._call_api(urls.hvac))

        if state and not state.is_pending:
            await self._call_api(urls.hvac_update, 'put')

    async def logout(self) -> None:
        """Get logged out from the API, see
        :func:`~pymultimatic.api.connector.ApiConnector.logout`
        """
        if not self._fixed_serial:
            self._serial = None
        await self._connector.logout()

    # pylint: disable=no-self-use
    def _round(self, number: float) -> float:
        """round a float to the nearest 0.5, as vaillant API only accepts 0.5
        step"""
        return round(number * 2) / 2

    async def _call_api(self,
                        url_call: Callable[..., str],
                        method: Optional[str] = None,
                        **kwargs: Any) -> Any:
        await self._ensure_ready()

        params = kwargs.get('params', {})
        params.update({'serial': self._serial})

        payload = kwargs.get('payload', None)

        if method is None:
            method = 'get'
            if payload is not None:
                method = 'put'

        url = url_call(**params)
        return await self._connector.request(method, url, payload)

    async def _ensure_ready(self) -> None:
        if not await self._connector.is_logged():
            await self._connector.login()
            await self._fetch_serial()
        if not self._serial:
            await self._fetch_serial()

    async def _fetch_serial(self) -> None:
        if not self._fixed_serial:
            facilities = await self._connector.get(urls.facilities_list())
            self._serial = mapper.map_serial_number(facilities)
